unit OperatorTimingTest;

{$mode objfpc}{$H+}
{$CODEALIGN LOCALMIN=16}
{$CODEALIGN CONSTMIN=16}

interface

uses
  Classes, SysUtils, fpcunit, testutils, testregistry,
  native, GLZVectorMath;

{$I config.inc}

type

  { TBaseTimingTest }

  TBaseTimingTest = class(TTestCase)
  protected
    procedure SetUp; override;
    procedure TearDown; override;
  public
    {$CODEALIGN RECORDMIN=16}
    nt1,nt2, nt3 : TNativeGLZVector4f;
    vt1,vt2, vt3 : TGLZVector4f;
    {$CODEALIGN RECORDMIN=4}
    r1: single;
    TestDispName: string;
    Res: boolean;
  end;

  { TFileWriteTest }

  TFileWriteTest = class (TTestCase)
  published
    procedure WriteFile;
    procedure WriteMarkDown;
    procedure WriteHTML;
    procedure ClearLog;

  end;

   { TOperatorTimingTest }

  TOperatorTimingTest = class(TBaseTimingTest)
  published
    procedure TimeAdd;
    procedure TimeSub;
    procedure TimeMul;
    procedure TimeDiv;
    procedure TimeAddSingle;
    procedure TimeSubSingle;
    procedure TimeMulSingle;
    procedure TimeDivSingle;
    procedure TimeNegative;
    procedure TimeOpEqual;
    procedure TimeOpGTEqual;
    procedure TimeOpLTEqual;
    procedure TimeOpGT;
    procedure TimeOpLT;
    procedure TimeOpNotEqual;
  end;

  { TOperationOnSelfTimingTest }

  TOperationOnSelfTimingTest = class(TBaseTimingTest)
  published
    procedure TimepAdd;
    procedure TimepSub;
    procedure TimepMul;
    procedure TimepDiv;
    procedure TimepAddSingle;
    procedure TimepSubSingle;
    procedure TimepMulSingle;
    procedure TimepDivSingle;
    procedure TimepInvert;
    procedure TimepNegate;
    procedure TimepAbs;
    procedure TimepNormalize;
    procedure TimepDivideBy2;
    procedure TimepCrossProduct;
    procedure TimeMinVector;
    procedure TimeMinSingle;
    procedure TimeMaxVector;
    procedure TimeMaxSingle;
  end;

const
  asmUnitVector : TGLZVector4f = (X:1; Y:1; Z:1; W:1);
  natUnitVector : TNativeGLZVector4f = (X:1; Y:1; Z:1; W:1);


{%region%====[ TBaseTimingTest etc ]===========================================}
var
  StartTime, TotalTime, elapsedTime : double;
  etNative, etAsm: double;
  cnt: Integer;
  sl,ml,hl: TStringList;

implementation

procedure InitLists;
begin
  sl := TStringList.Create;
  sl.Add('Compiler Flags: ' + REP_FLAGS);
  sl.Add('Test,Native,Assembler');

  ml := TStringList.Create;
  ml.Add('| Compiler Flags | ' + REP_FLAGS + ' |');
  ml.Add('| -------- | --------- |');
  ml.Add('');
  ml.Add('| Test  | Native | Assembler |');
  ml.Add('| -------- | -------- | --------- |');

  hl := TStringList.Create;
  hl.Add('[b]Compiler Flags: ' + REP_FLAGS + '[/b]' );
  hl.Add('[table]');
  hl.Add('[tr][td]Test[/td][td]Native[/td][td]Assembler[/td][/tr]');

end;

procedure StartTimer;
begin
  StartTime := 0;
  StartTime := now;
End;

procedure StopTimer;
begin
  elapsedTime := 0;
  elapsedTime :=(Now() - StartTime) *24 * 60 * 60;
End;

function WriteTimer:String;
begin
  Result := FloatToStr(elapsedTime)+' seconds';
  TotalTime := TotalTime + elapsedTime;
End;

function WriteTime(et:double):String;
begin
  Result := FloatToStrF(et,ffFixed,5,6);
  TotalTime := TotalTime + et;
End;

{ TFileWriteTest }

procedure TFileWriteTest.WriteFile;
begin
  sl.SaveToFile(REP_FILE_CSV);
end;

procedure TFileWriteTest.WriteMarkDown;
begin
  ml.SaveToFile(REP_FILE_MD);
end;

procedure TFileWriteTest.WriteHTML;
begin
  hl.Add('[/table]');
  hl.SaveToFile(REP_FILE_HTML);
end;

procedure TFileWriteTest.ClearLog;
begin
  sl.Free;
  ml.free;
  hl.free;
  InitLists;
end;

procedure TBaseTimingTest.SetUp;
begin
  nt1.Create(5.850,-15.480,8.512,1.5);
  nt2.Create(1.558,6.512,4.525,1.0);
  vt1.Create(5.850,-15.480,8.512,1.5);
  vt2.Create(1.558,6.512,4.525,1.0);
  r1 := 1.2345;
end;

procedure TBaseTimingTest.TearDown;
begin
  inherited TearDown;
  Sl.Add(TestDispName + ', ' + WriteTime(etNative) + ', ' + WriteTime(etAsm));
  ml.Add( '| ' + TestDispName + ' | ' + WriteTime(etNative)  + ' | ' +  WriteTime(etAsm) + ' | ');
  hl.Add('[tr][td]' + TestDispName + '[/td][td]' + WriteTime(etNative)  + '[/td][td]' +  WriteTime(etAsm) + '[/td][/tr]');
end;

{%endregion}
{%region%====[ TOperatorTimingTest ]===========================================}
procedure TOperatorTimingTest.TimeAdd;
begin
  TestDispName := 'Vector Op Add Vector';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt3 := nt1 + nt2; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt3 := vt1 + vt2; end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperatorTimingTest.TimeSub;
begin
  TestDispName := 'Vector Op Subtract Vector';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt3 := nt1 - nt2; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt3 := vt1 - vt2; end;
  StopTimer;
  etAsm := elapsedTime;
end;
   procedure TOperatorTimingTest.TimeMul;
begin
  TestDispName := 'Vector Op Multiply Vector';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt3 := nt1 * nt2; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt3 := vt1 * vt2; end;
  StopTimer;
  etAsm := elapsedTime;
end;
   procedure TOperatorTimingTest.TimeDiv;
begin
  TestDispName := 'Vector Op Divide Vector';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt3 := nt1 / nt2; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt3 := vt1 / vt2; end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperatorTimingTest.TimeAddSingle;
begin
  TestDispName := 'Vector Op Add Single';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt3 := nt1 + r1; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt3 := vt1 + r1; end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperatorTimingTest.TimeSubSingle;
begin
  TestDispName := 'Vector Op Subtract Single';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt3 := nt1 - r1; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt3 := vt1 - r1; end;
  StopTimer;
  etAsm := elapsedTime;
end;
procedure TOperatorTimingTest.TimeMulSingle;
begin
  TestDispName := 'Vector Op Multiply Single';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt3 := nt1 * r1; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt3 := vt1 * r1; end;
  StopTimer;
  etAsm := elapsedTime;
end;
procedure TOperatorTimingTest.TimeDivSingle;
begin
  TestDispName := 'Vector Op Divide Single';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt3 := nt1 / r1; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt3 := vt1 / r1; end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperatorTimingTest.TimeNegative;
begin
  TestDispName := 'Vector Op Negative';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt3 := -nt1; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt3 := -vt1; end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperatorTimingTest.TimeOpEqual;
begin
  TestDispName := 'Vector Op Equal';
  StartTimer;
  for cnt := 1 to 20000000 do begin res := nt1 = nt2; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin res := vt1 = vt2; end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperatorTimingTest.TimeOpGTEqual;
begin
  TestDispName := 'Vector Op GT or Equal';
  StartTimer;
  for cnt := 1 to 20000000 do begin res := nt1 >= nt2; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin res := vt1 >= vt2; end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperatorTimingTest.TimeOpLTEqual;
begin
  TestDispName := 'Vector Op LT or Equal';
  StartTimer;
  for cnt := 1 to 20000000 do begin res := nt1 <= nt2; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin res := vt1 <= vt2; end;
  StopTimer;
  etAsm := elapsedTime;

end;

procedure TOperatorTimingTest.TimeOpGT;
begin
  TestDispName := 'Vector Op Greater';
  StartTimer;
  for cnt := 1 to 20000000 do begin res := nt1 > nt2; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin res := vt1 > vt2; end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperatorTimingTest.TimeOpLT;
begin
  TestDispName := 'Vector Op Less';
  StartTimer;
  for cnt := 1 to 20000000 do begin res := nt1 < nt2; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin res := vt1 < vt2; end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperatorTimingTest.TimeOpNotEqual;
begin
  TestDispName := 'Vector Op Not Equal';
  StartTimer;
  for cnt := 1 to 20000000 do begin res := nt1 <> nt2; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin res := vt1 <> vt2; end;
  StopTimer;
  etAsm := elapsedTime;
end;
{%endregion%}

{%region%====[ TOperationOnSelfTimingTest ]====================================}

procedure TOperationOnSelfTimingTest.TimepAdd;
begin
  TestDispName := 'Add Vector To Self';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pAdd(nt2); end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.padd(vt2); end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperationOnSelfTimingTest.TimepSub;
begin
  TestDispName := 'Sub Vector from Self';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pSub(nt2); end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.pSub(vt2); end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperationOnSelfTimingTest.TimepMul;
begin
  TestDispName := 'Multiply Vector with Self';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pMul(natUnitVector); end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.pMul(asmUnitVector); end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperationOnSelfTimingTest.TimepDiv;
begin
  TestDispName := 'Divide Self by Vector';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pDiv(natUnitVector); end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.pDiv(asmUnitVector); end;
  StopTimer;
  etAsm := elapsedTime;

end;

procedure TOperationOnSelfTimingTest.TimepAddSingle;
begin
  TestDispName := 'Add Single To Self';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pAdd(0.0001); end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.padd(0.0001); end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperationOnSelfTimingTest.TimepSubSingle;
begin
  TestDispName := 'Sub Single from Self';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pSub(0.0001); end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.pSub(0.0001); end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperationOnSelfTimingTest.TimepMulSingle;
begin
  TestDispName := 'Multiply Self with single';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pMul(1.0); end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.pMul(1.0); end;
  StopTimer;
  etAsm := elapsedTime;
end;

// note in operations on self use 1 to avoid overflow etc
procedure TOperationOnSelfTimingTest.TimepDivSingle;
begin
  TestDispName := 'Divide Self by single';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pDiv(1.0); end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.pDiv(1.0); end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperationOnSelfTimingTest.TimepInvert;
begin
  TestDispName := 'Invert Self';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pInvert; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.pInvert; end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperationOnSelfTimingTest.TimepNegate;
begin
  TestDispName := 'Negate Self';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pNegate; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.pNegate; end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperationOnSelfTimingTest.TimepAbs;
begin
  TestDispName := 'Self Abs';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pAbs; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.pAbs; end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperationOnSelfTimingTest.TimepNormalize;
begin
  TestDispName := 'Self Normalize';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pNormalize; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.pNormalize; end;
  StopTimer;
  etAsm := elapsedTime;
end;

//have to a bit more to stop overflows/underflows
procedure TOperationOnSelfTimingTest.TimepDivideBy2;
begin
  TestDispName := 'Self Divideby2';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pAdd(1); nt1.DivideBy2; end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.pAdd(1); vt1.DivideBy2; end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperationOnSelfTimingTest.TimepCrossProduct;
begin
  TestDispName := 'Self CrossProduct Vector';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pCrossProduct(nt2); end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.pCrossProduct(vt2); end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperationOnSelfTimingTest.TimeMinVector;
begin
  TestDispName := 'Self Min Vector';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pMin(nt2); end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.pMin(vt2); end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperationOnSelfTimingTest.TimeMinSingle;
begin
  TestDispName := 'Self Min Single';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pMin(r1); end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.pMin(r1); end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperationOnSelfTimingTest.TimeMaxVector;
begin
  TestDispName := 'Self Max Vector';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pMax(nt2); end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.pMax(vt2); end;
  StopTimer;
  etAsm := elapsedTime;
end;

procedure TOperationOnSelfTimingTest.TimeMaxSingle;
begin
  TestDispName := 'Self Max Single';
  StartTimer;
  for cnt := 1 to 20000000 do begin nt1.pMax(r1); end;
  StopTimer;
  etNative := elapsedTime;
  StartTimer;
  For cnt:= 1 to 20000000 do begin vt1.pMax(r1); end;
  StopTimer;
  etAsm := elapsedTime;
end;

{%endregion}



initialization
  InitLists;
  RegisterTest(TOperatorTimingTest);
  RegisterTest(TOperationOnSelfTimingTest);
  RegisterTest(TFileWriteTest ); // must always be last in list!!!


Finalization
  sl.Free;
  hl.Free;
  ml.Free;
end.

